(* Type représentant un trie.
   Pour ce type et la fonction find, voir:

    https://gist.github.com/Tchou/c0c305621bc909e189db90821d78eb6f
*)

type 'a trie =
    Node of ('a option * (char * 'a trie) list)
  (** [empty] est le trie vide. *)
let empty = Node (None, [])

(** [is_empty t] est vrai si et seulement si [t] est vide. *)
let is_empty t =
  match t with
    Node (None, []) -> true
  | _ -> false

(** [find_trie key t] renvoie le (sous)-trie accessible
    dans [t] en suivant la clé [key].
    Lève l'exception [Not_found] si [key] n'est pas dans [t] *)
let find_trie key t =
  let rec find_node i t =
    if i = String.length key then t
    else
      match t with Node (_, l) -> find_list i l

  and find_list i l =
    let ci = key.[i] in
    match l with
      [] -> raise Not_found
    | (d, t) :: ll ->
      if ci > d then find_list i ll
      else if ci = d then find_node (i+1) t
      else raise Not_found
  in
  find_node 0 t

(** [find key t] renvoie la valeur accessible dans
    [t] en suivant la clé [key]. Lève l'exception [Not_found]
    si [key] n'est pas dans [t]. *)
let find key t =
  match find_trie key t with
    Node (Some v, _) -> v
  | Node (None, _) -> raise Not_found

(** Q1.1 On suit la même structure que find. *)
let add key v t =
  let rec insert_node i t =
    match t with
    | Node (o, l) ->
      if i = String.length key then Node (Some v, l)
      else Node (o, insert_list i l)
  and insert_list i l =
    let ci = key.[i] in
    match l with
    | [] -> [ (ci, insert_node (i + 1) empty) ]
    | (d, t) :: ll ->
      if ci > d then (d, t) :: insert_list i ll
      else if ci = d then (ci, insert_node (i + 1) t) :: ll
      else (ci, insert_node (i + 1) empty) :: l (* attention, ici l pour laisser (d,t)*)
  in
  insert_node 0 t
(* Q1.2 Test : *)
(* On essaye d'écrire un test un peu générique *)
let add_from_list l t =
  List.fold_left (fun acc (k, v) -> add k v acc) t l

let test2_2 () =
  (* Une list des valeurs de test *)
  let l = [ "a", 10; "abc", 11; "abd", 12 ] in
  (* Remplissage du dico à partir de la liste *)
  let dico =  add_from_list l empty in
  (* On parcours la liste et on s'assure que la valeur extraite
     du dico pour cette clé est la valeur initialement associée.
  *)
  List.iter (fun (k, v) -> assert (find k dico = v)) l

let () = test2_2 () (* ne doit pas provoquer d'erreur *)


(* Q1.3 Ici, pas besoin de faire des fonctions imbriquées car par
   besoin de paramètre auxiliaire à initialiser: *)
let rec union f t1 t2 =
  match (t1, t2) with
  | Node (None, l1), Node (None, l2) -> Node (None, union_list f l1 l2)

  | Node (Some v, l1), Node (None, l2) (* attention 2 cas en 1 *)
  | Node (None, l1), Node (Some v, l2) -> Node (Some v, union_list f l1 l2)

  | Node (Some v1, l1), Node (Some v2, l2) -> Node (Some (f (v1, v2)), union_list f l1 l2)
and union_list f l1 l2 =
  match (l1, l2) with
  | [], l
  | l, [] ->
    l
  | (c1, t1) :: ll1, (c2, t2) :: ll2 ->
    if c1 < c2 then (c1, t1) :: union_list f ll1 l2
    else if c1 > c2 then (c2, t2) :: union_list f l1 ll2
    else (c1, union f t1 t2) :: union_list f ll1 ll2

(* Q1.4 *)
let test2_4 () =
  let keys1 = [ "A", 1; "B", 1; "ACAC", 1; "BB", 1 ] in
  let keys2 = [ "", 1; "ABBA", 1; "B", 1; "BA", 1] in
  let t_add = add_from_list keys2 (add_from_list keys1 empty) in
  let t_union = union snd (add_from_list keys1 empty)
      (add_from_list keys2 empty)
  in
  assert (t_add = t_union)

let () =  test2_4 ()
(* Q1.5
   Si on réfléchit, il n'y a qu'un chemin possible pour une clé,
   quelle que soit la façon de l'insérer et donc, pour un ensemble
   de clé donné, un seul trie possible.

   Ce n'est pas le cas des tables de hachages ou des AVL, qui peuvent
   avoir des formes différentes en fonction de l'ordre d'insertion et du
   moment où ont été fait les redimentionnement ou ré-équilibrages.

*)

(* Q2.1 Construit une chaîne de caractères à partir d'une liste
   de caractères donnés à l'envers et sa longueur
*)


let rev_implode l =
  String.concat "" (List.rev_map (String.make 1) l)

(* (String.make 1) == (fun c -> String.make 1 c)
   On utilise l'application partielle pour avoir
   un code plus concis.
   Pour le reste:
   - on transforme chaque caractère en une chaîne de taille 1
   - on colle toute les chaînes ensemble sans séparateur.
*)

(* Q2.2 Première solution, adaptation de ce qui a été vu en cours,
   parcours préfixe classique. On accumule dans une liste
   qu'il faut renverser à la fin. *)
let all_keys t =
  let rec loop t path acc =
    match t with
      Node (Some _, l ) ->
      (* On est sur un nœud terminal, on ajoute la clé constituée
         de tous les caractères rencontrés jusqu'à ce nœud,
         puis on continue sur la liste des caractères en dessous. *)
      loop_list l path ((rev_implode path) :: acc)
    | Node (None, l) -> (* nœud non terminal, on continue simplement *)
      loop_list l path acc
  and loop_list l path acc =
    match l with
      [] -> acc
    | (c, t) :: ll ->
      let acc2 = loop t (c :: path) acc in
      loop_list ll path acc2
  in
  List.rev (loop t [] [])


(* Q2.3
   Il suffit de combiner find_trie et all_keys en faisant
   attention aux exceptions levées par find_trie.
*)
let complete prefix t =
  try
    (* on trouve le trie accessible depuis
       prefix *)
    let t = find_trie prefix t in
    (* on renvoie toutes les clésde ce trie,
       qui sont toutes les continuations possibles de 'prefix' *)
    all_keys t
  with Not_found -> []

(* Q2.4  Le petit programme, on quitte simplement avec ctrl-c *)
let complete_french () =
  let rec load_dico cin dico =
    match In_channel.input_line cin with
      None -> dico
    | Some s ->
      let ndico = add s true dico in (* on ajoute une constante bidon comme valeur. *)
      load_dico cin ndico
  in
  let cin = open_in "french.txt" in
  let dico = load_dico cin empty in
  let () = close_in cin in
  while true do
    Printf.printf "Entrée: \n";
    let s = read_line () in
    let s = String.lowercase_ascii s in
    let l = complete s dico in
    List.iter (fun word -> Printf.printf "> %s|%s\n" s word) l;
  done

let () =  (* complete_french *) ()

(** En dessous c'est bonus. Ça peut être intéressant, mais uniquement
    si tout le code précédent est parfaitement compris. *)


(* Exo 3 *)

module StrMap = Map.Make(String)

let fill_map l =
  List.fold_left (fun acc s ->StrMap.add s 1 acc) StrMap.empty l

let fill_hash l =
  let h = Hashtbl.create 16 in
  List.iter (fun s -> Hashtbl.replace h s 1) l;
  h

let fill_trie l =
  List.fold_left (fun acc s ->add s 1 acc) empty l

let limit = 1000000
let find_hash h s =
  for _i = 1 to limit do
    ignore (Hashtbl.find h s);
  done

let find_map m s =
  for _i = 1 to limit do
    ignore (StrMap.find s m);
  done

let find_trie t s =
  for _i = 1 to limit do
    ignore (find s t);
  done



let time f arg msg =
  let () = Gc.full_major () in
  let () = Gc.compact () in
  let t0 = Unix.gettimeofday () in
  let res = f arg in
  let t1 = Unix.gettimeofday () in
  Printf.printf "%s: %fms\n" msg ((t1 -. t0) *. 1000.);
  res


let test_speed () =
  let rec load_text cin acc =
    match In_channel.input_line cin with
      None -> acc
    | Some s ->
      let nacc = List.rev_append (String.split_on_char ' ' s) acc
      in load_text cin nacc
  in
  let l = load_text (open_in "french.txt") [] in
  let h = time fill_hash l "Fill Hashtable" in
  let m = time fill_map l "Fill Map" in
  let t = time fill_trie l "Fill Trie" in
  let _ = time (find_hash h) "programmation" "Find Hashtable" in
  let _ = time (find_map m) "programmation" "Find Map" in
  let _ = time (find_trie t) "programmation" "Find Trie" in
  ()


let () = test_speed ()

let cyan = "\x1b[96m"
let reset = "\x1b[0m"

module Simple =
struct
  (** Affichage simple. 
      2 difficultés
      (1) On imprime ligne par ligne (avec Printf.printf par exemple)
          Si on imprime récursivement, lorsqu'on imprime un nœud
          par exemple le nœud b (i) il faut penser à imprimer autant
          de barres verticales qu'il y a de chemins "passant devant"
          On peut faire ça en gardant une pile des fins de listes
          qui restent à imprimer.
          Lorsqu'on explore le premier (a, ...) :: ll, on ajoute
          ce ll à une pile que l'on maintient.
          Lorsqu'on arrive devant le b (i), pour chaque liste non
          vide de la pile, on affiche "| ". Pour une liste vide
          par exemple devant le nœud b (ii) qui est le dernier de
          sa liste, il y a [] dans la pile, on affiche "  "

          ┬a┬a─a
          │ ├b  <--(i)
          │ └c
          └b┬c  <--(ii)
            └d

      (2) Pour afficher une liste de couples (c, t) dans un nœud,
      il faut traiter spécialement le premier nœud de la liste,
      qui est affiché horizontalement par rapport à son parent
      précédé de ─ s'il est seul et ┬ s'il y a d'autres nœuds en dessous.
      On utilise pour ça une fonction intermédiaire qui appelle ensuite
      la fonction récursive sur les listes.
  *)
  (* affiche la marge avant un nœud.
     stack est la liste des fin de liste qu'il reste
     à afficher, et indique donc (une fois renversée),
     combien de chemin verticaux passent avant le nœud
     en cours d'affichage. *)
  let pr_margin stack =
    let l = List.rev stack in
    List.iter
      (fun r ->
         match r with
         | [] -> Printf.printf "  "
         | _ -> Printf.printf "│ ")
      l

  (* On affiche un caractère. On prend en argument le trie
     associé, s'il est terminal, le caractère est affiché
     en bleu.
     On prend aussi en argument le symbole de lien a afficher devant
     le caractère ("─", "┬", "└" ou "├")
     Petite subtilité, si le caractère est plus petit que ' '
     (caractère de controle) ou >= 128 (caractère unicode représenté
     sur plusieurs octets) on remplace par un . pour ne pas
     décaler l'affichage. )*)
  let pr_char symb c t =
    Printf.printf "%s" symb;
    let c = if Char.code c >= 128 || c < ' ' then '.' else c in
    match t with
    | Node (Some _, _) -> Printf.printf "%s%c%s" cyan c reset
    | _ -> Printf.printf "%c" c


  (* Draw trie demande juste l'affichage de la liste.
     dans draw_list, single est le symbole
     a afficher s'il n'y a pas de suivant et many
     celui a afficher s'il y a des suivants.
     Pour le premier (c, t) de la liste ces symboles sont "─" "┬",
     pour les autre, les symboles sont  "└" "├"
  *)
  let rec draw_trie p t =
    let (Node (_, l)) = t in
    draw_first p l
  and draw_first p l =
    match l with
      [] -> Printf.printf "\n"
    | (c, t) :: [] ->
      pr_char "─" c t;
      draw_trie ([] :: p) t
    | (c, t) :: l ->
      pr_char "┬" c t;
      draw_trie (l :: p) t;
      draw_list p l
  and draw_list p l =
    pr_margin p;  (* on affiche la marge devant les autres nœuds *)
    match l with
    | [] -> Printf.printf "\n"
    | [ (c, t) ] ->
      pr_char "└" c t;
      draw_trie ([] :: p) t;
    | (c, t) :: ll ->
      pr_char "├" c t;
      draw_trie (ll :: p) t;
      draw_list p ll

  let test_draw () =
    let keys1 = [ "A", 1; "B", 1; "ACAC", 1; "BB", 1 ] in
    let keys2 = [ "", 1; "ABBA", 1; "B", 1; "BA", 1] in
    let t_add = add_from_list keys2 (add_from_list keys1 empty) in
    let t_union = union snd (add_from_list keys1 empty)
        (add_from_list keys2 empty)
    in
    (* on peut constater que les 2 tries sont les mêmes*)
    draw_trie [] t_add;
    Printf.printf "-------\n";
    draw_trie [] t_union

  let () = test_draw ()
end